"""Minimalist implementation of Single Transferable Vote."""
import collections
import enum
import itertools

import numpy as np

from .types import PyStvError, RaceResult, RoundResult


class RoundMode(enum.Enum):
    """Method used to round votes_need = (# votes) / (# candidates + 1)"""

    CEILING = 1  # Used in test of FairVote data, though technically incorrect.
    ADD_ONE = 2  # Very close to correct, and easy to understand.
    FRACTIONAL = 3  # Technically most correct, but leads to a lot of fractional votes.


EPSILON = 1e-5


def run_stv(race_info, round_mode=RoundMode.ADD_ONE):
    num_cands = len(race_info.metadata.names)
    ballots = validate_and_standardize_ballots(race_info.ballots, num_cands)
    votes = np.array(race_info.votes)[:, None]

    num_slots = num_cands + 1  # zero is reserved for empty marks

    weights = np.zeros_like(ballots, float)
    weights[:, 0] = 1

    votes_needed = np.sum(votes) / (race_info.metadata.num_seats + 1)
    if round_mode == RoundMode.CEILING:
        votes_needed = np.ceil(votes_needed)
    elif round_mode == RoundMode.ADD_ONE:
        votes_needed = np.floor(1 + votes_needed)
    elif round_mode == RoundMode.FRACTIONAL:
        votes_needed += EPSILON
    else:
        raise ValueError(f"round_mode is not a value in RoundMode enum: {round_mode}")

    status = np.zeros(num_slots, int)
    status[0] = -2
    valid = np.ones_like(ballots, bool)
    orig = first_nonzero(valid, axis=1)  # Original candidate on each ballot

    round_info = []
    while np.count_nonzero(status > 0) < race_info.metadata.num_seats:
        counts = np.bincount(ballots.ravel(), (weights * votes).ravel(), num_slots)
        counts_masked = np.ma.array(counts, mask=(status != 0))
        elected_mask = counts_masked >= votes_needed

        elected = []
        eliminated = []

        if np.count_nonzero(status >= 0) == race_info.metadata.num_seats:
            to_remove = []
            elected = np.nonzero(status == 0)[0]
            multipliers = []
            status[status == 0] = 1
        elif elected_mask.any():
            to_remove = np.nonzero(elected_mask)[0]
            elected = to_remove
            multipliers = [votes_needed / counts_masked[c] for c in to_remove]
            status[to_remove] = 1
        else:  # eliminate candidate with fewest votes (random selection if >1)
            min_counts = np.where(counts_masked == counts_masked.min())[0]
            to_remove = [np.random.choice(min_counts)]
            eliminated = to_remove
            multipliers = [0]
            status[to_remove] = -1

        # Distribute votes from elected/eliminated candidates to the next active
        # lower ranked candidate.  Elected candidates transfer excess votes, while
        # eliminated candidates transfer all votes.
        transfers = {}
        for cand, mult in zip(to_remove, multipliers):
            valid[ballots == cand] = False
            next = first_nonzero(valid, axis=1)  # Next active candidate on each ballot

            # Ballots where the candidate changed
            o_rows = orig != next
            # Ballots the candidate changed, but where no active candidate remains
            n_rows = o_rows & (next != -1)

            # Votes are transferred to next candidate.
            weights[n_rows, next[n_rows]] = weights[n_rows, orig[n_rows]] * (1 - mult)
            # Votes are transferred from original candidate.
            weights[o_rows, orig[o_rows]] *= mult

            transfers_cand = collections.defaultdict(int)
            slicer = n_rows, next[n_rows]
            for c, v in zip(ballots[slicer], (weights * votes)[slicer]):
                if v:
                    transfers_cand[c] += v

            if transfers_cand:
                transfers[cand] = dict(transfers_cand)
            orig = next

        if not np.isclose(counts.sum(), votes.sum()):
            raise PyStvError("Final round count total does not equal original votes")

        round_info.append(
            RoundResult(counts.tolist(), list(elected), list(eliminated), transfers)
        )
    return RaceResult(race_info.metadata, round_info)


def first_nonzero(arr, axis, invalid_val=-1):
    mask = arr != 0
    return np.where(mask.any(axis=axis), mask.argmax(axis=axis), invalid_val)


def validate_and_standardize_ballots(ballots, num_cands):
    # Fill out ballots to be the same length.
    ballots = np.array(list(itertools.zip_longest(*ballots, fillvalue=0))).T
    if ballots.ndim != 2:
        raise ValueError("Ballot data has wrong dim: %s" % ballots.ndim)

    check_oob(ballots, num_cands)

    # Check duplicates
    np.apply_along_axis(check_dup_1d, axis=1, arr=ballots)

    # Add a 0 at the end of every ballot.
    return np.pad(ballots, ((0, 0), (0, 1)))


def check_oob(ballots, num_cands):
    ballots = np.asarray(ballots)
    oob_rankings = (ballots < 0) | (ballots > num_cands)
    if oob_rankings.any():
        bad_ballots = oob_rankings.any(axis=1)
        bad_indices = np.nonzero(bad_ballots)[0].tolist()
        raise ValueError(
            f"Ranking out of allowed range on ballots indices: {bad_indices}"
        )


def check_dup_1d(row):
    unique, counts = np.unique(row, return_counts=True)
    has_dup = (counts[unique != 0] > 1).any()
    if has_dup:
        raise PyStvError(f"Ballot has duplicated entry: {row}")
    return has_dup
