"""Common types used in pystv."""
from dataclasses import dataclass

from icontract import invariant


class PyStvError(Exception):
    """Error in PyStv."""


@dataclass
class RaceMetadata:
    race_name: str
    num_seats: int
    names: list[str]

    def __str__(self):
        cands = ",".join(self.names)
        return (
            f"race: {self.race_name}\nnum_seats: {self.num_seats}\ncandidates: {cands}"
        )


@invariant(lambda self: len(self.metadata.names) >= max(map(max, self.ballots)))
@invariant(lambda self: len(self.ballots) == len(self.votes))
@dataclass
class RaceInfo:
    metadata: RaceMetadata
    ballots: list[list[int]]
    votes: list


@dataclass
class RoundResult:
    count: list[float]
    elected: list[int]
    eliminated: list[int]
    transfers: dict[int, dict[int, float]]


@invariant(lambda self: len(set(len(r.count) for r in self.rounds)) < 2)
@invariant(
    lambda self: not self.rounds
    or (len(self.rounds[0].count) == len(self.metadata.names) + 1)
)
@dataclass
class RaceResult:
    metadata: RaceMetadata
    rounds: list[RoundResult]

    def __str__(self):
        def rnd2str(round_result):
            details = []
            names_all = ["<exhausted>"] + self.metadata.names
            for idx, (name, cnt) in enumerate(zip(names_all, round_result.count)):
                name_detail = f"{name}: {cnt}"
                if idx in round_result.elected:
                    name_detail += " +"
                if idx in round_result.eliminated:
                    name_detail += " -"
                details.append(name_detail)
                # Add transfers
            return "\n".join(details)

        round_info = []
        for rnd, round_results in enumerate(self.rounds):
            round_info.extend([f"Round {rnd}:", rnd2str(round_results)])
        rounds = "\n".join(round_info)
        return f"{self.metadata}\n{rounds}"
